#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#if defined(HAS_TCP) || defined(HAS_TERM)
#ifdef HAS_NET_ERRNO_H
#include <sys/bsdtypes.h>
#include <net/errno.h>
#endif
#endif
#ifdef HAS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifndef QDOS
#include <sys/socket.h>
#else
#include <qdos.h>
#include <qptr.h>
extern long qlchan;
#endif
#include <fcntl.h>

#ifdef NOISEDEBUG
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#endif

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "lutil.h"
#include "ttyio.h"

#define PUSHMAX 8

extern int hanged_up;

int timed_out;
int tty_status=0;
int f_flags;

#ifdef NOISEDEBUG
extern int junklevel;
extern void damage(char*,int*);
#endif

static int waitmode=1;
static int tty_wait(void);
static int tty_nowait(void);

#define bufsiz 4096
static char buf[bufsiz];
static char *next;
static int left=0;

static char pushbuf[PUSHMAX];
static int pushindex;

static void almhdl(int sig)
{
        debug(17,"timeout");
        timed_out=1;
}

static void get_buf(int tot)
{
#ifndef QDOS
        if (tty_status&STAT_HANGUP) return;
        timed_out=0;
        signal(SIGALRM,almhdl);
        alarm(tot);
#endif
#ifdef QDOS
    timeout_t qwait;
    int res,xtra;

    timed_out=0;
    qwait = tot*50*waitmode;
    
    /* just try for a single char within timeout (worst case unix 'success') */

    res = io_fbyte(qlchan, qwait, buf);
    if(res == 0)
    {
        left = 1;
        tty_status = 0;

        /*  see if any 'bonus' characters but don't hang around looking */
        /*  or maybe with just a very small t/o (between 1 - 5 clocks)  */
 
        xtra  = io_fstrg(qlchan, 0, buf+1, bufsiz-1);
        if(xtra > 0)
        {
            left += xtra;
        }
    }
    else
    {
        switch (res)
        {
          case -1:
            timed_out = 1;
            tty_status = waitmode ? STAT_TIMEOUT : STAT_EMPTY;
            break;
          case -6:
          case -10:
            tty_status=STAT_HANGUP;
            break;
        }
    }
    if (tty_status & STAT_ERROR) logerr("get_buf err %d, left %d, wait %d",
                                              res,left,waitmode);
    if (tty_status) left=0;
    next=buf;
/*    debug(18,"get_buf returning %d bytes, status=0x%02x",left,tty_status); */
#else
        left=read(0,buf,bufsiz);
        if (left < 0) err=errno;
        else err=0;
#ifdef NOISEDEBUG
        damage(buf,&left);
#endif
        alarm(0);
        signal(SIGALRM,SIG_DFL);
        if (hanged_up)
                tty_status=STAT_HANGUP;
        else if (timed_out)
                tty_status=STAT_TIMEOUT;
        else if (left == 0)
                tty_status=STAT_EMPTY;
        else if (left < 0)
        {
                if ((err == 0) || (err == EAGAIN))
                        tty_status=STAT_EMPTY;
                else if ((err == EPIPE)
#if defined(HAS_TCP) || defined(HAS_TERM)
                                        || (err == ECONNRESET)
#endif
                                                              )
                        tty_status=STAT_HANGUP;
                else
                        tty_status=STAT_ERROR;
        }
        else
                tty_status=0;
        if (tty_status & STAT_ERROR)
                logerr("$get_buf error");
        if (tty_status) left=0;
        next=buf;
        debug(18,"get_buf returning %d bytes, status=0x%02x",left,tty_status);
#endif
}

int tty_ungetc(int c)
{
        if (pushindex >= PUSHMAX) return -1;
        pushbuf[pushindex++]=c;
        return 0;
}

int tty_getc(int tot)
{
        int c;

        if (pushindex) return pushbuf[--pushindex]&0xff;

        if (!left) get_buf(tot);
        if (tty_status) c=-tty_status;
        else
        {
                left--;
                c=(*(next++))&0xff;
        }
/*        if(c >= 0) debug(18,"getc return '%s'",printablec(c));
        else debug(18,"getc return %d",c);
*/
        return c;
}

int tty_get(char *str,int l,int tot)
{
        int got,err;
#ifdef QDOS
        timeout_t qwait;
        int res,xtra;
#endif
        if (pushindex >= l)
        {
                while (l)
                {
                        *(str++)=pushbuf[--pushindex];
                        l--;
                }
                return 0;
        }

        while (pushindex)
        {
                *(str++)=pushbuf[--pushindex];
                l--;
        }

        if (left >= l)
        {
                memcpy(str,next,l);
                next+=l;
                left-=l;
                return 0;
        }

        if (left > 0)
        {
                memcpy(str,next,left);
                str+=left;
                next+=left;
                l-=left;
                left=0;
        }

        if (tty_status&STAT_HANGUP) return tty_status;

        timed_out=0;
#ifndef QDOS
        signal(SIGALRM,almhdl);
        alarm(tot);
#endif
#ifdef QDOS
        qwait = tot*50*waitmode;

        while (l && (tty_status == 0))
        {
    /* just try for a single char within timeout (worst case unix 'success') */

                 res = io_fbyte(qlchan, qwait, str);
                 if(res == 0)
                 {
                     got = 1;
                     tty_status = 0;

     /*  see if any 'bonus' characters but don't hang around looking */
     /*  or maybe with just a very small t/o (between 1 - 5 clocks)  */

                     xtra  = io_fstrg(qlchan, 0, str+1, l-1);
                     if(xtra > 0)
                     {
                         got += xtra;
                     }
                 }
                 else
                 {
                     switch (res)
                     {
                       case -1:
                         timed_out = 1;
                         tty_status= waitmode ? STAT_TIMEOUT : STAT_EMPTY;
                         break;
                       case -6:
                       case -10:
                         tty_status=STAT_HANGUP;
                         break;
                    }
                 }
                 if (tty_status & STAT_ERROR)
                        logerr("tty_get err %d, got %d, wait %d",
                                                        err,got,waitmode);
                 if (tty_status) got=0;
                 if(got > 0)
                 {
                   str+=got;
                   l-=got;
                 }
        }
        debug(18,"tty_get %d bytes missing, status=0x%02x",l,tty_status);
        return tty_status;
#else   /* not qdos */
        while (l && (tty_status == 0))
        {
                got=read(0,str,l);
                if (got < 0) err=errno;
                else err=0;
#ifdef NOISEDEBUG
                damage(str,&got);
#endif
                if (hanged_up)
                        tty_status=STAT_HANGUP;
                else if (timed_out)
                        tty_status=STAT_TIMEOUT;
                else if (got == 0)
                        tty_status=STAT_EMPTY;
                if (got < 0)
                {
                        if ((err == 0) || (err == EAGAIN))
                                tty_status |= STAT_EMPTY;
                        else if ((err == EPIPE)
#if defined(HAS_TCP) || defined(HAS_TERM)
                                                || (err == ECONNRESET)
#endif
                                                                      )
                                tty_status |= STAT_HANGUP;
                        else
                                tty_status=STAT_ERROR;
                }
                else
                        tty_status=0;
                if (tty_status & STAT_ERROR) logerr("$tty_get error");
                if(got > 0)
                {
                  str+=got;
                  l-=got;
                }
        }
        alarm(0);
        signal(SIGALRM,SIG_DFL);
        debug(18,"tty_get %d bytes missing, status=0x%02x",l,tty_status);
        return tty_status;
#endif
}

int tty_putc(int c)
{
        char buf;
        int  ferr;

        if (tty_status&STAT_HANGUP) return -tty_status;
        buf=c&0xff;
        tty_status=0;
#ifdef QDOS
        ferr = io_sbyte(qlchan,-1,buf);
        if(ferr != 0)
#else
        ttywait(1);
        if (write(1,&buf,1) != 1)
#endif
        {
                if ((hanged_up) ||
                    (errno == EPIPE)
#if defined(HAS_TCP) || defined(HAS_TERM)
                                     || (errno == ECONNRESET)
#endif
                                                            )
                        tty_status=STAT_HANGUP;
                else tty_status=STAT_ERROR;
        }
        if (tty_status) logerr("tty_putc status=0x%02x",tty_status);
        else debug(18,"putc '%s'",printablec(c));
        return -tty_status;
}

int tty_put(char *str,int l)
{
        int written;

        if (tty_status&STAT_HANGUP) return -tty_status;
        tty_status=0;
        if (l <= 0) return 0;
#ifdef QDOS
        written = io_sstrg(qlchan,-1,str,l);
        if(written != l)
#else
        ttywait(1);
        if ((written=write(1,str,l)) != l)
#endif
        {
                if ((hanged_up) ||
                    (errno == EPIPE)
#if defined(HAS_TCP) || defined(HAS_TERM)
                                     || (errno == ECONNRESET)
#endif
                                                             )
                        tty_status=STAT_HANGUP;
                else tty_status=STAT_ERROR;
        }
        if (tty_status)
        {
                int tmpfl=fcntl(1,F_GETFL,0L);
                logerr("tty_put status=0x%02x (req %d, wrote %d, fl 0%x)",
                                tty_status,l,written,tmpfl);
        }
        else debug(18,"tty_put %d bytes",l);
        return -tty_status;
}


int ttywait(int on)
{
        int rc=0;

        if (on == waitmode) return 0;
#ifdef QDOS
        waitmode = on;
        return 0;
#else
        debug(18,"setting port waitmode %s",on?"on":"off");
        if (on) rc=tty_wait();
        else rc=tty_nowait();
        if (rc == 0) waitmode=on;
        return rc;
#endif
}

int tty_wait(void)
{
        int rc;

        if ((rc=fcntl(0,F_SETFL,f_flags & ~O_NONBLOCK)))
        {
                logerr("SETFL normal ioctl error");
        }
        return rc;
}

int tty_nowait(void)
{
        int rc;

        if ((rc=fcntl(0,F_SETFL,f_flags | O_NONBLOCK)))
        {
                logerr("SETFL nonblock ioctl error");
        }
        return rc;
}

#ifdef NOISEDEBUG

#ifndef RAND_MAX
#ifdef INT_MAX
#define RAND_MAX INT_MAX
#else
#define RAND_MAX 2147483647
#endif
#endif

void damage(char *buf,int *len)
{
        long width;
        int i;

        if ((junklevel == 0) || (*len == 0)) return;

        width=RAND_MAX/junklevel;
        if (rand() < width)
        {
                if (rand() & 4)
                {
                        debug(18,"length %d damaged - decreased (if > 1)",*len);
                        if ((*len) > 1) (*len)--;
                }
                else
                {
                        debug(18,"length %d damaged - increased",*len);
                        (*len)++;
                }
        }
        else for (i=0;i<*len;i++) if (rand() < width)
        {
                debug(18,"char %d of %d damaged",i,*len);
                buf[i]=rand();
        }
}

#endif
